package _VisualDVM.GlobalData;
import Common.CommonConstants;
import Common.Database.VisualiserDatabase;
import Common.MainModule_;
import _VisualDVM.Global;
import _VisualDVM.GlobalData.Compiler.Compiler;
import _VisualDVM.GlobalData.Compiler.CompilersDBTable;
import _VisualDVM.GlobalData.DBLastProject.LastProjectsDBTable;
import _VisualDVM.GlobalData.DVMParameter.DVMParameterDBTable;
import _VisualDVM.GlobalData.EnvironmentValue.EnvironmentValuesDBTable;
import _VisualDVM.GlobalData.Machine.Machine;
import _VisualDVM.GlobalData.Machine.MachinesDBTable;
import _VisualDVM.GlobalData.Makefile.MakefilesDBTable;
import _VisualDVM.GlobalData.Module.ModulesDBTable;
import _VisualDVM.GlobalData.RunConfiguration.RunConfigurationsDBTable;
import _VisualDVM.GlobalData.SapforProfile.SapforProfile;
import _VisualDVM.GlobalData.SapforProfile.SapforProfilesDBTable;
import _VisualDVM.GlobalData.SapforProfileSetting.SapforProfileSetting;
import _VisualDVM.GlobalData.SapforProfileSetting.SapforProfileSettingsDBTable;
import _VisualDVM.GlobalData.Tasks.CompilationTask.CompilationTasksDBTable;
import _VisualDVM.GlobalData.Tasks.RunTask.RunTasksDBTable;
import _VisualDVM.GlobalData.User.User;
import _VisualDVM.GlobalData.User.UsersDBTable;
import _VisualDVM.Passes.PassCode;

import java.nio.file.Paths;
import java.util.Date;
import java.util.LinkedHashMap;
public class GlobalDatabase extends VisualiserDatabase {
    //---------СЕАНС----------------------------------------------
    public MachinesDBTable machines;
    public UsersDBTable users;
    public CompilersDBTable compilers;
    public MakefilesDBTable makefiles;
    public ModulesDBTable modules;
    public CompilationTasksDBTable compilationTasks;
    public RunTasksDBTable runTasks;
    public RunConfigurationsDBTable runConfigurations;
    public EnvironmentValuesDBTable environmentValues;
    public DVMParameterDBTable dvmParameters;
    //----- ДАННЫЕ ВИЗУАЛИЗАТОРА---------------------------------
    public LastProjectsDBTable lastProjects;
    //-
    public SapforProfilesDBTable sapforProfiles = null;
    //---------
    public SapforProfileSettingsDBTable sapforProfilesSettings = null;
    //-
    public GlobalDatabase() {
        super(Paths.get(System.getProperty("user.dir"), "Data", "db7.sqlite").toFile());
    }
    @Override
    protected void initAllTables() throws Exception {
        super.initAllTables();
        addTable(machines = new MachinesDBTable());
        addTable(users = new UsersDBTable());
        addTable(compilers = new CompilersDBTable());
        addTable(makefiles = new MakefilesDBTable());
        addTable(modules = new ModulesDBTable());
        addTable(compilationTasks = new CompilationTasksDBTable());
        addTable(runTasks = new RunTasksDBTable());
        addTable(runConfigurations = new RunConfigurationsDBTable());
        addTable(environmentValues = new EnvironmentValuesDBTable());
        addTable(lastProjects = new LastProjectsDBTable());
        addTable(dvmParameters = new DVMParameterDBTable());
        addTable(sapforProfiles = new SapforProfilesDBTable());
        addTable(sapforProfilesSettings = new SapforProfileSettingsDBTable());
    }
    //---------------------------------------------------------------------------------
    @Override
    public void Init() throws Exception {
        //настройки компонент
        runConfigurations.Patch();
        //---
        MainModule_.instance.getUI().getFastAccessMenuBar().Refresh();
    }
    @Override
    public PassCode getSynchronizePassCode() {
        return null;
    }
    //todo централизованное сохранение текущих объектов в базах
    public void SaveCredentials() {
        Global.normalProperties.updateField("lastMachineId", getTable(Machine.class).getUI().getCurrentPK(CommonConstants.Nan));
        Global.normalProperties.updateField("lastUserId", getTable(User.class).getUI().getCurrentPK(CommonConstants.Nan));
        Global.normalProperties.updateField("lastCompilerId", getTable(Compiler.class).getUI().getCurrentPK(CommonConstants.Nan));
    }
    public void RestoreCredentials() {
        if (Global.normalProperties.lastMachineId != CommonConstants.Nan) {
            getTable(Machine.class).getUI().SetCurrentByPK(Global.normalProperties.lastMachineId);
            //-
            if (Global.normalProperties.lastUserId != CommonConstants.Nan)
                getTable(User.class).getUI().SetCurrentByPK(Global.normalProperties.lastUserId);
            //--
            if (Global.normalProperties.lastCompilerId != CommonConstants.Nan)
                getTable(Compiler.class).getUI().SetCurrentByPK(Global.normalProperties.lastCompilerId);
        }
    }
    //--
    //проверить, есть ли профиль с таким же набором настроек.
    public SapforProfile checkProfileForCurrentSettings() {
        LinkedHashMap<String, String> current_values = Global.mainModule.getProject().sapforProperties.toMap();
        for (SapforProfile profile : sapforProfiles.Data.values()) {
            //--получить все настройки профиля
            LinkedHashMap<String, String> profileValues = new LinkedHashMap<>();
            for (SapforProfileSetting setting : sapforProfilesSettings.Data.values())
                if (setting.sapforprofile_id == profile.id) profileValues.put(setting.name, setting.value);
            //--
            if (current_values.equals(profileValues)) return profile;
        }
        return null;
    }
    public void insertProfileSettings(SapforProfile profile) throws Exception {
        LinkedHashMap<String, String> current_values = Global.mainModule.getProject().sapforProperties.toMap();
        for (String name : current_values.keySet()) {
            //--
            SapforProfileSetting sapforProfileSetting = new SapforProfileSetting();
            sapforProfileSetting.name = name;
            sapforProfileSetting.value = current_values.get(name);
            sapforProfileSetting.sapforprofile_id = profile.id;
            //--
            Insert(sapforProfileSetting);
        }
    }
    public void saveCurrentProfile(String name_in) throws Exception {
        if (checkProfileForCurrentSettings() == null) {
            SapforProfile profile = new SapforProfile();
            profile.description = name_in;
            profile.creationDate = new Date().getTime();
            Insert(profile);
            insertProfileSettings(profile);
        }
    }
    public void rewriteProfileByDescription(String description_in) throws Exception {
        SapforProfile old = null;
        //ищем профиль с именем бага
        for (SapforProfile p : sapforProfiles.Data.values())
            if (p.description.equalsIgnoreCase(description_in)) {
                old = p;
                break;
            }
        if (old != null) {
            //удалить профиль с именем бага,если есть, чтобы не множить кроликов.
            //при условии что не видно окна профилей!!
            Delete(old);
            DeleteByFK(old, SapforProfileSetting.class);
        }
        //---
        SapforProfile profile = new SapforProfile();
        profile.description = description_in;
        profile.creationDate = new Date().getTime();
        Insert(profile);
        insertProfileSettings(profile);
    }
    public SapforProfile getDefaultSapforProfile() {
        for (SapforProfile sapforProfile : sapforProfiles.Data.values()) {
            if (sapforProfile.isAuto != 0)
                return sapforProfile;
        }
        return null;
    }
}
